package ru.dbotthepony.mc.otm.datagen.tags

import it.unimi.dsi.fastutil.objects.ObjectArraySet
import net.minecraft.core.HolderLookup
import net.minecraft.core.Registry
import net.minecraft.core.registries.BuiltInRegistries
import net.minecraft.core.registries.Registries
import net.minecraft.resources.ResourceKey
import net.minecraft.resources.ResourceLocation
import net.minecraft.tags.BlockTags
import net.minecraft.tags.GameEventTags
import net.minecraft.tags.TagKey
import net.minecraft.world.item.Item
import net.minecraft.world.item.Tier
import net.minecraft.world.item.Tiers
import net.minecraft.world.level.block.Block
import net.minecraftforge.data.event.GatherDataEvent
import net.minecraftforge.registries.ForgeRegistries
import net.minecraftforge.registries.IForgeRegistry
import ru.dbotthepony.mc.otm.capability.MatteryPlayerCapability
import ru.dbotthepony.mc.otm.datagen.DataGen
import java.util.stream.Stream
import net.minecraft.data.tags.TagsProvider as MinecraftTagsProvider

interface MTagAppender<T> {
	fun add(value: T): MTagAppender<T>
	fun addSafe(value: T): Boolean
	fun add(vararg values: T): MTagAppender<T>
	fun addSafe(vararg values: T): Boolean {
		var any = false
		for (value in values) any = addSafe(value)
		return any
	}

	fun add(values: Collection<T>): MTagAppender<T>
	fun add(values: Stream<out T>): MTagAppender<T> = add(values.toList())

	fun addSafe(values: Collection<T>): Boolean {
		var any = false
		for (value in values) any = addSafe(value)
		return any
	}
}

interface ForgeTagAppender<T> : MTagAppender<T> {
	fun add(key: String, value: T): ForgeTagAppender<T>
	fun addSafe(key: String, value: T): Boolean
	fun add(key: String, vararg values: T): ForgeTagAppender<T>
	fun addSafe(key: String, vararg values: T): Boolean {
		var any = false
		for (value in values) any = addSafe(key, value)
		return any
	}

	fun add(key: String, values: Collection<T>): ForgeTagAppender<T>

	fun addSafe(key: String, values: Collection<T>): Boolean {
		var any = false
		for (value in values) any = addSafe(key, value)
		return any
	}
}

private fun <T> vanillaLookup(key: ResourceKey<Registry<T>>): (T) -> ResourceLocation? {
	val registry by lazy { (BuiltInRegistries.REGISTRY.get(key.location()) ?: throw NoSuchElementException("No such registry $key")) as Registry<T> }
	return { registry.getKey(it) }
}

class TagsProvider(
	private val event: GatherDataEvent
) {
	inner class Delegate<T> private constructor(key: ResourceKey<Registry<T>>, val lookup: (T) -> ResourceLocation?) : MinecraftTagsProvider<T>(event.generator.packOutput, key, event.lookupProvider, DataGen.MOD_ID, event.existingFileHelper) {
		constructor(registry: IForgeRegistry<T>) : this(registry.registryKey, registry::getKey)
		constructor(key: ResourceKey<Registry<T>>) : this(key, vanillaLookup(key))

		init {
			event.generator.addProvider(true, this)
		}

		private val tags = HashMap<TagKey<T>, ObjectArraySet<T>>()

		override fun addTags(provider: HolderLookup.Provider) {
			if (tags.isEmpty()) {
				return
			}

			for ((tag, values) in tags) {
				tag(tag).also {
					for (value in values) {
						it.add(ResourceKey.create(registryKey, lookup(value) ?: throw NoSuchElementException("$value is missing from $registryKey")))
					}
				}
			}
		}

		fun getSet(tag: TagKey<T>): MutableSet<T> {
			return tags.computeIfAbsent(tag) { ObjectArraySet() }
		}

		fun getSet(location: ResourceLocation) = getSet(TagKey.create(registryKey, location))

		fun appender(tag: TagKey<T>, message: ((T) -> Any) = { "$it is already in $tag" }): MTagAppender<T> {
			val list = getSet(tag)

			return object : MTagAppender<T> {
				override fun add(value: T): MTagAppender<T> {
					if (!list.add(value)) {
						throw IllegalStateException(message.invoke(value).toString())
					}

					return this
				}

				override fun add(values: Collection<T>): MTagAppender<T> {
					for (value in values) add(value)
					return this
				}

				override fun addSafe(value: T): Boolean {
					return list.add(value)
				}

				override fun add(vararg values: T): MTagAppender<T> {
					values.forEach(this::add)
					return this
				}
			}
		}

		fun forge(path: String): ForgeTagAppender<T> {
			val parent by lazy { appender(ResourceLocation("forge", path)) }

			return object : ForgeTagAppender<T> {
				override fun add(key: String, value: T): ForgeTagAppender<T> {
					val tag = TagKey.create(registryKey, ResourceLocation("forge", "$path/$key"))

					if (!getSet(tag).add(value)) {
						throw IllegalStateException("$value is already in $tag")
					}

					return this
				}

				override fun addSafe(key: String, value: T): Boolean {
					val tag = TagKey.create(registryKey, ResourceLocation("forge", "$path/$key"))
					return getSet(tag).add(value)
				}

				override fun add(key: String, vararg values: T): ForgeTagAppender<T> {
					for (value in values) add(key, value)
					return this
				}

				override fun add(key: String, values: Collection<T>): ForgeTagAppender<T> {
					for (value in values) add(key, value)
					return this
				}

				override fun add(value: T): MTagAppender<T> {
					return parent.add(value)
				}

				override fun addSafe(value: T): Boolean {
					return parent.addSafe(value)
				}

				override fun add(vararg values: T): MTagAppender<T> {
					return parent.add(*values)
				}

				override fun add(values: Collection<T>): MTagAppender<T> {
					return parent.add(values)
				}
			}
		}

		fun appender(location: ResourceLocation): MTagAppender<T> {
			return appender(TagKey.create(registryKey, location))
		}

		fun appender(location: ResourceLocation, message: (T) -> Any): MTagAppender<T> {
			return appender(TagKey.create(registryKey, location), message)
		}
	}

	val blocks = Delegate(ForgeRegistries.BLOCKS)
	val items = Delegate(ForgeRegistries.ITEMS)
	val mobEffects = Delegate(ForgeRegistries.MOB_EFFECTS)

	val androidImmuneEffects = mobEffects.appender(MatteryPlayerCapability.ANDROID_IMMUNE_EFFECTS)

	val requiresShovel = blocks.appender(BlockTags.MINEABLE_WITH_SHOVEL)
	val requiresAxe = blocks.appender(BlockTags.MINEABLE_WITH_AXE)
	val requiresHoe = blocks.appender(BlockTags.MINEABLE_WITH_HOE)
	val requiresPickaxe = blocks.appender(BlockTags.MINEABLE_WITH_PICKAXE)

	val requiresStoneTool = blocks.appender(BlockTags.NEEDS_STONE_TOOL)
	val requiresIronTool = blocks.appender(BlockTags.NEEDS_IRON_TOOL)
	val requiresDiamondTool = blocks.appender(BlockTags.NEEDS_DIAMOND_TOOL)

	val witherImmune = blocks.appender(BlockTags.WITHER_IMMUNE)
	val dragonImmune = blocks.appender(BlockTags.DRAGON_IMMUNE)

	fun stoneOre(key: String, block: Block): TagsProvider {
		ore(key, block)

		itemGroundOresStone.add(block.asItem())
		blockGroundOresStone.add(block)

		return this
	}

	fun deepslateOre(key: String, block: Block): TagsProvider {
		ore(key, block)

		itemGroundOresDeepslate.add(block.asItem())
		blockGroundOresDeepslate.add(block)

		return this
	}

	fun singleDropOre(vararg blocks: Block): TagsProvider {
		for (block in blocks) {
			itemOreRatesSingular.add(block.asItem())
			blockOreRatesSingular.add(block)
		}

		return this
	}

	fun ore(key: String, block: Block): TagsProvider {
		val forgeKey = ResourceLocation("forge", "ores/$key")
		val b = TagKey.create(Registries.BLOCK, forgeKey)
		val i = TagKey.create(Registries.ITEM, forgeKey)

		items.getSet(i).add(block.asItem())
		itemOres.add(block.asItem())
		blocks.getSet(b).add(block)
		blockOres.add(block)

		return this
	}

	fun clump(key: String, clump: Item): TagsProvider {
		rawMaterials.add(key, clump)
		rawMaterials.add(clump)
		return this
	}

	fun clump(key: String, clump: Item, rawBlock: Block): TagsProvider {
		val rawKey = "raw_$key"

		rawMaterials.add(key, clump)
		rawMaterials.add(clump)
		storageBlocksAsItem.add(rawKey, rawBlock.asItem())
		storageBlocksAsItem.add(rawBlock.asItem())
		storageBlocksAsBlock.add(rawKey, rawBlock)
		storageBlocksAsBlock.add(rawBlock)

		return this
	}

	val circuits = items.forge("circuits")
	val dusts = items.forge("dusts")
	val ingots = items.forge("ingots")
	val itemOres = items.forge("ores")
	val blockOres = blocks.forge("ores")
	val plates = items.forge("plates")
	val storageBlocksAsItem = items.forge("storage_blocks")
	val storageBlocksAsBlock = blocks.forge("storage_blocks")
	val rawMaterials = items.forge("raw_materials")
	val wires = items.forge("wires")

	val blockGroundOresStone = blocks.forge("ores_in_ground/stone")
	val blockGroundOresDeepslate = blocks.forge("ores_in_ground/deepslate")
//	val blockGroundOresNetherrack = blocks.forge("ores_in_ground/netherrack")
	val itemGroundOresStone = items.forge("ores_in_ground/stone")
	val itemGroundOresDeepslate = items.forge("ores_in_ground/deepslate")
//	val itemGroundOresNetherrack = items.forge("ores_in_ground/netherrack")

//	val blockOreRatesSparse = blocks.forge("ore_rates/sparse")
	val blockOreRatesSingular = blocks.forge("ore_rates/singular")
//	val blockOreRatesDense = blocks.forge("ore_rates/dense")
//	val itemOreRatesSparse = items.forge("ore_rates/sparse")
	val itemOreRatesSingular = items.forge("ore_rates/singular")
//	val itemOreRatesDense = items.forge("ore_rates/dense")

	val gameEvents = Delegate(Registries.GAME_EVENT)
	val vibrations = gameEvents.appender(GameEventTags.VIBRATIONS)

	fun requiresPickaxe(block: Block, tier: Tier? = null): TagsProvider {
		requiresPickaxe.add(block)

		when (tier) {
			Tiers.STONE -> requiresStoneTool.add(block)
			Tiers.IRON -> requiresIronTool.add(block)
			Tiers.DIAMOND, Tiers.NETHERITE -> requiresDiamondTool.add(block)
		}

		return this
	}

	fun requiresPickaxe(blocks: Collection<Block>, tier: Tier? = null): TagsProvider {
		for (block in blocks) requiresPickaxe(block, tier)
		return this
	}

	fun requiresPickaxe(vararg blocks: Block, tier: Tier? = null): TagsProvider {
		for (block in blocks) requiresPickaxe(block, tier)
		return this
	}

	fun requiresShovel(block: Block, tier: Tier? = null): TagsProvider {
		requiresShovel.add(block)

		when (tier) {
			Tiers.STONE -> requiresStoneTool.add(block)
			Tiers.IRON -> requiresIronTool.add(block)
			Tiers.DIAMOND, Tiers.NETHERITE -> requiresDiamondTool.add(block)
		}

		return this
	}

	fun requiresShovel(blocks: Collection<Block>, tier: Tier? = null): TagsProvider {
		for (block in blocks) requiresShovel(block, tier)
		return this
	}

	fun requiresAxe(block: Block, tier: Tier? = null): TagsProvider {
		requiresAxe.add(block)

		when (tier) {
			Tiers.STONE -> requiresStoneTool.add(block)
			Tiers.IRON -> requiresIronTool.add(block)
			Tiers.DIAMOND, Tiers.NETHERITE -> requiresDiamondTool.add(block)
		}

		return this
	}

	fun requiresAxe(blocks: Collection<Block>, tier: Tier? = null): TagsProvider {
		for (block in blocks) requiresAxe(block, tier)
		return this
	}
}

